# a port of Peter Elsea's fuzzy functions for LISP
# by Oeyvind Brandtsegg 2005 - obrandts@online.no

import itertools
import math


# basic functions
def fzWeight(list, value):
    """Multiplies all memebers of alist by value"""
    outlist = []
    for item in list:
        item = item * value
        outlist.append(item)
    return outlist

def fzComplement (alist, compvalue=1.0):
    """returns fuzzy complement (1 - membership) of a list"""
    outlist=[]
    for item in alist:
        outlist.append(compvalue - item)
    return outlist 
 
def fzIntersection (*lists):
    """Returns fuzzy intersection (minimums) of any number of lists"""
    outlist = []
    for item in itertools.izip(*lists):
        outlist.append(min(item))
    return outlist

def fzInterpolate (list1, list2, weight=0.5):
    """Interpolates between two lists,
    optional weight parameter gives more or less weight to one of the lists.
    A weight of 0 yields no weight to list1 and full weight to list2,
    a weight of 0.5 gives equal weight to both lists"""
    outlist = []
    for item1, item2 in itertools.izip(fzWeight(list1, 1-weight),
                                       fzWeight(list2, weight)):
        outlist.append(item1+item2)
    return outlist

def fzUnion (*lists):
    """Returns fuzzy union (maximums) of any number of lists"""
    outlist = []
    for item in itertools.izip(*lists):
        outlist.append(max(item))
    return outlist


###################
# Fuzzy Inference

def fzFind(list, target):
    """returns fractional position of target in list"""
    old = list[0]
    # exception handling, if target value is not fractional member of list,
    # use max or min as new target
    if target > max(list):
        target = max(list)
        print "fzFind: target out of bounds, uses max as target"
    if target < min(list):
        target = min(list)
        print "fzFind: target out of bounds, uses min as target"
    for index,item in enumerate(list):
        if item > target > old:
            return index + ((target - old) / (item - old)) - 1
        if item < target < old:
            return index + ((old - target) / (old - item))- 1
        if item == target:
            return index
        old = item

def fzMembership(list, index):
    """returns the interpolated membership at a fractional location in list.
    Out of range indices are clipped to valid indices"""
    if index < 0 :
        index = 0
        print "fzMembership: index below zero"
    if index > (len(list) -1) :
        index = len(list) -1
        print "fzMembership: index higher than list length"
    value1 = list[int(math.floor(index))]
    value2 = list[int(math.ceil(index))]
    fract = index - math.floor(index)
    return value1 + ((value2 - value1) * fract)

def fzRule(value, predicate, consequent):
    """the fuzzy way of saying if VALUE is PREDICATE then CONSEQUENT"""
    return fzWeight(consequent, fzMembership(predicate, value))
